//------------------------------------------------------------------------
//Metasequoia.cpp
//------------------------------------------------------------------------

#include "stdafx.h"
#include "VixarLibrary.h"
#include "Strings.h"


#define COMCHECK(i,s) afs->chkcom(s,i)
#define CMPSTR(i,s) afs->chkstr(i,s)

#define LOADERMQO_SUCCEEDED       0
#define LOADERMQO_ERROR          -1
#define LOADERMQO_EOF            -2
#define LOADERMQO_HEADER_ERR     -3
#define LOADERMQO_FORMAT_ERR     -4
#define LOADERMQO_TRIAL_NOISE    -5

//------------------------------------------------------------------------
//LoaderMQANX
//------------------------------------------------------------------------
class LoaderMQA
{
protected:
	//LN^[|C^
	CCharacter *character;
	//t@C|C^
	FILE *fp;
	//AXL[t@C
	AsciiFileString *afs;
	//}eA
	std::vector<CMaterial> materials;
	//Aj[VZbgp
	bool animationFlg;

protected:
	//wb_[`FbN
	BOOL CheckHeader(void);
	//s
	BOOL IgnoreUnknown(int string_count);
	//oCi
	BOOL IgnoreBinary(int string_count);
	//C[v
	BOOL MainLoop(void);
	//}eAǂݍ
	BOOL LoadMaterial(void);
	//IuWFNgǂݍ
	BOOL LoadObject(CModel* model);
	//_ǂݍ
	BOOL LoadVertex(CModel* model);
	//ʓǂݍ
	BOOL LoadFace(CModel* model);
	//Aj[Vǂݍ
	BOOL LoadAnimation(CModel* model);

public:
	//RXgN^
	LoaderMQA(CCharacter *pt,FILE *handle);
	//fXgN^
	virtual ~LoaderMQA();
	//ǂݍ
	int Load(void);
};

//------------------------------------------------------------------------
//MQAEMQOt@C̓ǂݍ
//------------------------------------------------------------------------
bool CCharacter::LoadMQAFile(std::string file_name,LPDIRECT3DDEVICE9 dev)
{
	int ret_code;
	std::ostringstream stream;
	stream << GetDirectory() << file_name;
	FILE *fp = fopen(stream.str().c_str(), "rb");
	if(fp == NULL) return NULL;

	LoaderMQA *loader = new LoaderMQA(this,fp);

	ret_code = loader->Load();

	delete loader;

	fclose(fp);

	SearchCollisionModel();
	LoadTexture(dev);
	CreateVertexBuffer(dev);

	return (ret_code == LOADERMQO_SUCCEEDED);
}

//------------------------------------------------------------------------
//RXgN^
//------------------------------------------------------------------------
LoaderMQA::LoaderMQA(CCharacter *pt,FILE *handle)
{
	character = pt;

	fp = handle;

	materials.clear();

	afs = new AsciiFileString(fp);

	animationFlg = false;
}

//------------------------------------------------------------------------
//fXgN^
//------------------------------------------------------------------------
LoaderMQA::~LoaderMQA()
{
	delete afs;
}

//------------------------------------------------------------------------
//wb_[`FbN
//------------------------------------------------------------------------
BOOL LoaderMQA::CheckHeader(void)
{
	int sc;

	//check header
	sc = afs->GetLine();
	if(!afs->chkcom("Metasequoia",2) || !afs->chkstr(1,"Document"))
	{
		throw(LOADERMQO_HEADER_ERR);
	}

	//check format
	sc = afs->GetLine();
	if(!afs->chkcom("Format",4) || !afs->chkstr(1,"Text")
	|| !afs->chkstr(2,"Ver") || afs->ToDouble(3)!=1.0)
	{
		throw(LOADERMQO_FORMAT_ERR);
	}

	return TRUE;
}

//------------------------------------------------------------------------
//oCi
//------------------------------------------------------------------------
BOOL LoaderMQA::IgnoreBinary(int string_count)
{
	long oldpos, newpos;
	long offset = afs->ToInt(string_count-2);

	oldpos = ftell(fp);
	fseek(fp, offset, SEEK_CUR);
	newpos = ftell(fp);

	if(newpos - oldpos != offset)
	{
		throw(LOADERMQO_ERROR);
	}

	return TRUE;
}

//------------------------------------------------------------------------
//s
//------------------------------------------------------------------------
BOOL LoaderMQA::IgnoreUnknown(int string_count)
{
	if(string_count >= 1 && CMPSTR(string_count-1,"{"))
	{   //pass unknown group
		for(int comc=1; comc>0; )
		{
			string_count = afs->GetLine();
			if(string_count == EOF)
			{
				throw(LOADERMQO_EOF);
			}
			else if(string_count >= 1)
			{
				if(CMPSTR(string_count-1,"{"))
				{
					comc++;
				}
				else if(CMPSTR(string_count-1,"}"))
				{
					comc--;
				}
				else if( string_count>=3
				&& afs->CheckString(string_count-3,"[")
				&& afs->CheckString(string_count-1,"]") )
				{
					IgnoreBinary(string_count-2);
				}
			}
		}
	}
	else if( string_count>=3 && afs->CheckString(string_count-3,"[") && afs->CheckString(string_count-1,"]") )
	{
		IgnoreBinary(string_count-2);
	}

	return TRUE;
}

//------------------------------------------------------------------------
//C[v
//------------------------------------------------------------------------
BOOL LoaderMQA::MainLoop(void)
{
	for(;;)
	{
		int sc = afs->GetLine();
		if(sc == EOF) break;
		if(sc == 0) continue;

		//gCAmCY
		if( afs->chkcom("TrialNoise",2) )
		{
			throw(LOADERMQO_ERROR);
		}
		else if( afs->chkcom("Material",2) )//}eA
		{
			if( !afs->chkstr(sc-1,"{") )
			{
				throw(LOADERMQO_ERROR);
			}
			if( !LoadMaterial() )
			{
				throw(LOADERMQO_ERROR);
			}
		}
		else if( afs->chkcom("Object",2) )//IuWFNg
		{
			if(!afs->chkstr(sc-1,"{"))
			{
				throw(LOADERMQO_ERROR);
			}

			CModel* model;
			int result = 0;

			// f1₷
			character->PushBackModel(CModel());
			model = character->GetModelsBack();

			try
			{
				if(sc >= 3)
				{
					//fZbg
					model->SetName(afs->GetString(1));
				}
				if( !LoadObject(model) )
				{
					throw(LOADERMQO_ERROR);
				}
			}
			catch(int err_code)
			{
				result = err_code;
			}
			if(result)
			{
				throw(result);
			}
		}
		else
		{
			IgnoreUnknown(sc);
		}
	}

	return TRUE;
}

//------------------------------------------------------------------------
//ǂݍ
//------------------------------------------------------------------------
int LoaderMQA::Load(void)
{
	try
	{
		if( !CheckHeader() )
		{
			throw(LOADERMQO_ERROR);
		}
		if( !MainLoop() )
		{
			throw(LOADERMQO_ERROR);
		}
	}
	catch(int err_code)
	{
		return err_code;
	}

	return LOADERMQO_SUCCEEDED;
}

//------------------------------------------------------------------------
//}eAǂݍ
//------------------------------------------------------------------------
BOOL LoaderMQA::LoadMaterial(void)
{
	int i,scn;

	for(;;)
	{
		CMaterial mat = CMaterial();

		int sc = afs->GetLine();
		if(sc == EOF)
		{
			throw(LOADERMQO_EOF);
		}
		if(sc == 0) continue;

		if(COMCHECK(1,"}")) break;

		mat.SetName(afs->GetString(0));

		for(scn=1; scn<sc;)
		{
			if(CMPSTR(scn,"col") && scn+7<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+6,")")) break;
				mat.SetRGBA(afs->ToFloat(scn+2),afs->ToFloat(scn+3),
							afs->ToFloat(scn+4),afs->ToFloat(scn+5));
				scn += 7;
			}
			else if(CMPSTR(scn,"shader") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat.SetShader(afs->ToInt(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"dif") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat.SetDiffuse(afs->ToFloat(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"amb") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat.SetAmbient(afs->ToFloat(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"spc") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat.SetSpecular(afs->ToFloat(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"emi") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat.SetEmissive(afs->ToFloat(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"power") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat.SetPower(afs->ToFloat(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"tex") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat.SetTextureName(afs->GetString(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"aplane") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat.SetAlphaName(afs->GetString(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"bump") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat.SetBumpName(afs->GetString(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn+1,"("))
			{
				for(i=2; i+scn<sc; i++)
				{
					if(CMPSTR(scn+i,")")) break;
				}
				scn += i;
			}
			else
			{
				break;
			}
		}

		mat.SetMaterial();
		character->PushBackMaterial(mat);
	}

	return TRUE;
}

//------------------------------------------------------------------------
//IuWFNgǂݍ
//------------------------------------------------------------------------
BOOL LoaderMQA::LoadObject(CModel* model)
{
	for(;;)
	{
		int sc = afs->GetLine();
		if(sc == EOF)
		{
			throw(LOADERMQO_EOF);
		}
		if(COMCHECK(1,"}"))
		{
			break;
		}
		else if(COMCHECK(2,"facet"))
		{
			model->SetFacet(afs->ToFloat(1));
		}
		else if(COMCHECK(2,"mirror_axis"))
		{
			model->SetMirrorAxis(afs->ToInt(1));
		}
		else if(COMCHECK(3,"vertex"))
		{
			if(!CMPSTR(sc-1,"{")) return FALSE;
			if(!LoadVertex(model)) return FALSE;
		}
		else if(COMCHECK(3,"face"))
		{
			if(!CMPSTR(sc-1,"{")) return FALSE;
			if(!LoadFace(model)) return FALSE;
		}
		else if(COMCHECK(3,"animation"))
		{
			if(!CMPSTR(sc-1,"{")) return FALSE;
			if(!LoadAnimation(model)) return FALSE;
		}
		else
		{
			IgnoreUnknown(sc);
		}
	}
	//OȂ̂2ڈȍ~̓Aj[VǉȂtO
	animationFlg = true;

	return TRUE;
}

//------------------------------------------------------------------------
//_ǂݍ
//------------------------------------------------------------------------
BOOL LoaderMQA::LoadVertex(CModel* model)
{
	for(;;)
	{
		int sc = afs->GetLine();
		if (sc == EOF)
		{
			throw(LOADERMQO_EOF);
		}
		if( COMCHECK(1,"}")) break;

		if(sc < 3) continue;

		model->PushBackVertex(D3DXVECTOR3(afs->ToFloat(0),afs->ToFloat(1),afs->ToFloat(2)));
	}

	return TRUE;
}

//------------------------------------------------------------------------
//ʓǂݍ
//------------------------------------------------------------------------
BOOL LoaderMQA::LoadFace(CModel* model)
{
	int scn;
	int vc;

	for(;;)
	{
		CFace face;
		int sc = afs->GetLine();
		if( sc == EOF )
		{
			throw(LOADERMQO_EOF);
		}
		if( sc == 0 ) continue;

		if ( COMCHECK(1,"}") ) break;

		vc = afs->ToInt(0); // vertex count
		scn = 1;

		if ( CMPSTR(scn,"V") && scn+vc+3 <= sc )
		{
			if ( !CMPSTR(scn+1,"(") || !CMPSTR(scn+vc+2,")") ) continue;
			if ( vc >= 1 ) face.PushBackIndex(afs->ToInt(scn+2));
			if ( vc >= 2 ) face.PushBackIndex(afs->ToInt(scn+3));
			if ( vc >= 3 ) face.PushBackIndex(afs->ToInt(scn+4));
			if ( vc >= 4 ) face.PushBackIndex(afs->ToInt(scn+5));
			scn += vc+3;
		}
		else throw(LOADERMQO_ERROR);

		if ( CMPSTR(scn,"M") && scn+4 <= sc )
		{
			if ( !CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")") ) continue;
			face.SetMaterial(afs->ToInt(scn+2));
			scn += 4;
		}
		if ( CMPSTR(scn,"UV") && scn+vc*2+3 <= sc )
		{
			if ( !CMPSTR(scn+1,"(") || !CMPSTR(scn+vc*2+2,")") ) continue;

			if ( vc >= 1 ) { face.PushBackUV(CUV(afs->ToFloat(scn+2),afs->ToFloat(scn+3))); }
			if ( vc >= 2 ) { face.PushBackUV(CUV(afs->ToFloat(scn+4),afs->ToFloat(scn+5))); }
			if ( vc >= 3 ) { face.PushBackUV(CUV(afs->ToFloat(scn+6),afs->ToFloat(scn+7))); }
			if ( vc >= 4 ) { face.PushBackUV(CUV(afs->ToFloat(scn+8),afs->ToFloat(scn+9))); }
			scn += vc*2+3;
		}

		model->PushBackFace(face);
	}

	return TRUE;
}

//------------------------------------------------------------------------
//Aj[Vǂݍ
//------------------------------------------------------------------------
BOOL LoaderMQA::LoadAnimation(CModel* model)
{
	model->PushBackAnimation();
	int i = model->GetAnimationSize()-1;

	if ( !animationFlg )
	{
		//Aj[VZbg
		character->PushBackAnimationName(afs->GetString(1));
	}

	for(;;)
	{
		int sc = afs->GetLine();
		if (sc == EOF)
		{
			throw(LOADERMQO_EOF);
		}
		if( COMCHECK(1,"}")) break;

		if(sc < 16) continue;

		//4s4Zbg
		D3DXMATRIX mat;

		mat._11 = afs->ToFloat(0);
		mat._12 = afs->ToFloat(1);
		mat._13 = afs->ToFloat(2);
		mat._14 = afs->ToFloat(3);

		mat._21 = afs->ToFloat(4);
		mat._22 = afs->ToFloat(5);
		mat._23 = afs->ToFloat(6);
		mat._24 = afs->ToFloat(7);

		mat._31 = afs->ToFloat(8);
		mat._32 = afs->ToFloat(9);
		mat._33 = afs->ToFloat(10);
		mat._34 = afs->ToFloat(11);

		mat._41 = afs->ToFloat(12);
		mat._42 = afs->ToFloat(13);
		mat._43 = afs->ToFloat(14);
		mat._44 = afs->ToFloat(15);

		model->PushBackAnimation(i,mat);
	}

	return TRUE;
}
